/* Copyright (C) 2012 to 2014 Chris Vine

The library comprised in this file or of which this file is part is
distributed by Chris Vine under the GNU Lesser General Public
License as follows:

   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public License
   as published by the Free Software Foundation; either version 2.1 of
   the License, or (at your option) any later version.

   This library is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License, version 2.1, for more details.

   You should have received a copy of the GNU Lesser General Public
   License, version 2.1, along with this library (see the file LGPL.TXT
   which came with this source code package in the c++-gtk-utils
   sub-directory); if not, write to the Free Software Foundation, Inc.,
   51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

However, it is not intended that the object code of a program whose
source code instantiates a template from this file or uses macros or
inline functions (of any length) should by reason only of that
instantiation or use be subject to the restrictions of use in the GNU
Lesser General Public License.  With that in mind, the words "and
macros, inline functions and instantiations of templates (of any
length)" shall be treated as substituted for the words "and small
macros and small inline functions (ten lines or less in length)" in
the fourth paragraph of section 5 of that licence.  This does not
affect any other reason why object code may be subject to the
restrictions in that licence (nor for the avoidance of doubt does it
affect the application of section 2 of that licence to modifications
of the source code in this file).

*/

#ifndef CGU_TASK_MANAGER_TPP
#define CGU_TASK_MANAGER_TPP

namespace Cgu {

namespace Thread {

namespace TaskManagerHelper {

class Empty {};

template<class Obj, class Func>
struct arg_result_0 {
  Obj obj;
  Func func;
  arg_result_0(Obj obj_, Func func_):
    obj(obj_), func(func_) {}
};

template<class Obj, class Func, class Arg1>
struct arg_result_1 {
  Obj obj;
  Func func;
  typename Cgu::RemoveRefCond<Arg1, true>::Type arg1;
  arg_result_1(Obj obj_, Func func_,
	       typename Cgu::Param<Arg1>::ParamType arg1_):
    obj(obj_), func(func_), arg1(arg1_) {}
};

template<class Obj, class Func, class Arg1, class Arg2>
struct arg_result_2 {
  Obj obj;
  Func func;
  typename Cgu::RemoveRefCond<Arg1, true>::Type arg1;
  typename Cgu::RemoveRefCond<Arg2, true>::Type arg2;
  arg_result_2(Obj obj_, Func func_,
	       typename Cgu::Param<Arg1>::ParamType arg1_,
	       typename Cgu::Param<Arg2>::ParamType arg2_):
    obj(obj_), func(func_), arg1(arg1_), arg2(arg2_) {}
};

// It is safe to hold 'do_fail_cb' by pointer, because that callback
// always needs to execute if something else fails by throwing and the
// implementation always deletes it when it has been finished with or
// has not been used.
template<class Obj, class Func>
struct arg_when_0 {
  Obj obj;
  Func func;
  const Callback::Callback* do_fail_cb;
  gint priority;
  GMainContext* context;
  arg_when_0(Obj obj_, Func func_,
	     const Callback::Callback* do_fail_cb_,
	     gint priority_, GMainContext* context_):
    obj(obj_), func(func_), do_fail_cb(do_fail_cb_),
    priority(priority_), context(context_) {}
};

// It is safe to hold 'do_fail_cb' by pointer, because that callback
// always needs to execute if something else fails by throwing and the
// implementation always deletes it when it has been finished with or
// has not been used.
template<class Obj, class Func, class Arg1>
struct arg_when_1 {
  Obj obj;
  Func func;
  const Callback::Callback* do_fail_cb;
  gint priority;
  GMainContext* context;
  typename Cgu::RemoveRefCond<Arg1, true>::Type arg1;
  arg_when_1(Obj obj_, Func func_,
	     const Callback::Callback* do_fail_cb_,
	     gint priority_, GMainContext* context_,
	     typename Cgu::Param<Arg1>::ParamType arg1_):
    obj(obj_), func(func_), do_fail_cb(do_fail_cb_),
    priority(priority_), context(context_), arg1(arg1_) {}
};

// It is safe to hold 'do_fail_cb' by pointer, because that callback
// always needs to execute if something else fails by throwing and the
// implementation always deletes it when it has been finished with or
// has not been used.
template<class Obj, class Func, class Arg1, class Arg2>
struct arg_when_2 {
  Obj obj;
  Func func;
  const Callback::Callback* do_fail_cb;
  gint priority;
  GMainContext* context;
  typename Cgu::RemoveRefCond<Arg1, true>::Type arg1;
  typename Cgu::RemoveRefCond<Arg2, true>::Type arg2;
  arg_when_2(Obj obj_, Func func_,
	     const Callback::Callback* do_fail_cb_,
	     gint priority_, GMainContext* context_,
	     typename Cgu::Param<Arg1>::ParamType arg1_,
	     typename Cgu::Param<Arg2>::ParamType arg2_):
    obj(obj_), func(func_), do_fail_cb(do_fail_cb_),
    priority(priority_), context(context_), arg1(arg1_), arg2(arg2_) {}
};

struct MemfunResultWrapper {
  template <class Obj, class Func, class Ret>
  static void exec0(const arg_result_0<Obj, Func>& args,
		    const SharedLockPtr<AsyncResult<Ret> >& ret) {
    ret->set((args.obj.*args.func)());
  }
  template <class Obj, class Func, class Ret, class Arg1>
  static void exec1(const arg_result_1<Obj, Func, Arg1>& args,
		    const SharedLockPtr<AsyncResult<Ret> >& ret) {
    ret->set((args.obj.*args.func)(args.arg1));
  }
  template <class Obj, class Func, class Ret, class Arg1, class Arg2>
  static void exec2(const arg_result_2<Obj, Func, Arg1, Arg2>& args,
		    const SharedLockPtr<AsyncResult<Ret> >& ret) {
    ret->set((args.obj.*args.func)(args.arg1, args.arg2));
  }
  template <class Ret>
  static void do_fail(const SharedLockPtr<AsyncResult<Ret> >& ret) {
    ret->set_error(); // won't throw
  }
};

struct FunResultWrapper {
  template <class Func, class Ret>
  static void exec0(const arg_result_0<Empty, Func>& args,
		    const SharedLockPtr<AsyncResult<Ret> >& ret) {
    ret->set(args.func());
  }
  template <class Func, class Ret, class Arg1>
  static void exec1(const arg_result_1<Empty, Func, Arg1>& args,
		    const SharedLockPtr<AsyncResult<Ret> >& ret) {
    ret->set(args.func(args.arg1));
  }
  template <class Func, class Ret, class Arg1, class Arg2>
  static void exec2(const arg_result_2<Empty, Func, Arg1, Arg2>& args,
		    const SharedLockPtr<AsyncResult<Ret> >& ret) {
    ret->set(args.func(args.arg1, args.arg2));
  }
  template <class Ret>
  static void do_fail(const SharedLockPtr<AsyncResult<Ret> >& ret) {
    ret->set_error(); // won't throw
  }
};

struct MemfunWhenWrapper {
  template <class Obj, class Func, class Ret>
  static void do_task0(const arg_when_0<Obj, Func>& args,
		       const Callback::SafeFunctorArg<const Ret&>& when) {
    Callback::post(Callback::make_ref(&do_when<Ret>,
                                      when,
                                      (args.obj.*args.func)()),
                   args.priority,
                   args.context);
    // if this callback doesn't throw so post_fail() doesn't execute,
    // we need to delete do_fail_cb.  Destructors of bound arguments
    // of do_fail_cb are not supposed to throw, but we might as well
    // deal with it here.  An elipsis catch block is fine as thread
    // cancellation is blocked in all tasks.
    try {
      delete args.do_fail_cb;
    }
    catch (...) {}
  }

  template <class Obj, class Func, class Ret, class Arg1>
  static void do_task1(const arg_when_1<Obj, Func, Arg1>& args,
		       const Callback::SafeFunctorArg<const Ret&>& when) {
    Callback::post(Callback::make_ref(&do_when<Ret>,
                                      when,
                                      (args.obj.*args.func)(args.arg1)),
                   args.priority,
                   args.context);
    // if this callback doesn't throw so post_fail() doesn't execute,
    // we need to delete do_fail_cb.  Destructors of bound arguments
    // of do_fail_cb are not supposed to throw, but we might as well
    // deal with it here.  An elipsis catch block is fine as thread
    // cancellation is blocked in all tasks.
    try {
      delete args.do_fail_cb;
    }
    catch (...) {}
  }

  template <class Obj, class Func, class Ret, class Arg1, class Arg2>
  static void do_task2(const arg_when_2<Obj, Func, Arg1, Arg2>& args,
		       const Callback::SafeFunctorArg<const Ret&>& when) {
    Callback::post(Callback::make_ref(&do_when<Ret>,
                                      when,
                                      (args.obj.*args.func)(args.arg1, args.arg2)),
                   args.priority,
                   args.context);
    // if this callback doesn't throw so post_fail() doesn't execute,
    // we need to delete do_fail_cb.  Destructors of bound arguments
    // of do_fail_cb are not supposed to throw, but we might as well
    // deal with it here.  An elipsis catch block is fine as thread
    // cancellation is blocked in all tasks.
    try {
      delete args.do_fail_cb;
    }
    catch (...) {}
  }

  template <class Ret>
  static void do_when(const Callback::SafeFunctorArg<const Ret&>& when,
                      const Ret& val) {
    when(val);
  }

  template <class Obj, class Func, class Ret>
  static void do_task_rel0(const arg_when_0<Obj, Func>& args,
			   const SharedLockPtr<SafeEmitterArg<const Ret&> >& when_emitter) {
    Callback::post(Callback::make_ref(&do_when_rel<Ret>,
                                      when_emitter,
                                      (args.obj.*args.func)()),
                   args.priority,
                   args.context);
    // if this callback doesn't throw so post_fail() doesn't execute,
    // we need to delete do_fail_cb.  Destructors of bound arguments
    // of do_fail_cb are not supposed to throw, but we might as well
    // deal with it here.  An elipsis catch block is fine as thread
    // cancellation is blocked in all tasks.
    try {
      delete args.do_fail_cb;
    }
    catch (...) {}
  }

  template <class Obj, class Func, class Ret, class Arg1>
  static void do_task_rel1(const arg_when_1<Obj, Func, Arg1>& args,
			   const SharedLockPtr<SafeEmitterArg<const Ret&> >& when_emitter) {
    Callback::post(Callback::make_ref(&do_when_rel<Ret>,
                                      when_emitter,
                                      (args.obj.*args.func)(args.arg1)),
                   args.priority,
                   args.context);
    // if this callback doesn't throw so post_fail() doesn't execute,
    // we need to delete do_fail_cb.  Destructors of bound arguments
    // of do_fail_cb are not supposed to throw, but we might as well
    // deal with it here.  An elipsis catch block is fine as thread
    // cancellation is blocked in all tasks.
    try {
      delete args.do_fail_cb;
    }
    catch (...) {}
  }

  template <class Obj, class Func, class Ret, class Arg1, class Arg2>
  static void do_task_rel2(const arg_when_2<Obj, Func, Arg1, Arg2>& args,
			   const SharedLockPtr<SafeEmitterArg<const Ret&> >& when_emitter) {
    Callback::post(Callback::make_ref(&do_when_rel<Ret>,
                                      when_emitter,
                                      (args.obj.*args.func)(args.arg1, args.arg2)),
                   args.priority,
                   args.context);
    // if this callback doesn't throw so post_fail() doesn't execute,
    // we need to delete do_fail_cb.  Destructors of bound arguments
    // of do_fail_cb are not supposed to throw, but we might as well
    // deal with it here.  An elipsis catch block is fine as thread
    // cancellation is blocked in all tasks.
    try {
      delete args.do_fail_cb;
    }
    catch (...) {}
  }

  template <class Ret>
  static void do_when_rel(const SharedLockPtr<SafeEmitterArg<const Ret&> >& when_emitter,
                          const Ret& val) {
    when_emitter->emit(val);
  }

  static void post_fail(const Callback::Callback* do_fail_cb,
                        std::pair<gint, GMainContext*> loop_args) {
    // this function does not throw
    Callback::post(do_fail_cb, loop_args.first, loop_args.second);
  }

  static void do_fail_rel(const SharedLockPtr<SafeEmitter>& fail_emitter) {
    fail_emitter->emit(); // might throw std::bad_alloc
  }
};

struct FunWhenWrapper {
  template <class Func, class Ret>
  static void do_task0(const arg_when_0<Empty, Func>& args,
		       const Callback::SafeFunctorArg<const Ret&>& when) {
    Callback::post(Callback::make_ref(&do_when<Ret>,
                                      when,
                                      args.func()),
                   args.priority,
                   args.context);
    // if this callback doesn't throw so post_fail() doesn't execute,
    // we need to delete do_fail_cb.  Destructors of bound arguments
    // of do_fail_cb are not supposed to throw, but we might as well
    // deal with it here.  An elipsis catch block is fine as thread
    // cancellation is blocked in all tasks.
    try {
      delete args.do_fail_cb;
    }
    catch (...) {}
  }

  template <class Func, class Ret, class Arg1>
  static void do_task1(const arg_when_1<Empty, Func, Arg1>& args,
		       const Callback::SafeFunctorArg<const Ret&>& when) {
    Callback::post(Callback::make_ref(&do_when<Ret>,
                                      when,
                                      args.func(args.arg1)),
                   args.priority,
                   args.context);
    // if this callback doesn't throw so post_fail() doesn't execute,
    // we need to delete do_fail_cb.  Destructors of bound arguments
    // of do_fail_cb are not supposed to throw, but we might as well
    // deal with it here.  An elipsis catch block is fine as thread
    // cancellation is blocked in all tasks.
    try {
      delete args.do_fail_cb;
    }
    catch (...) {}
  }

  template <class Func, class Ret, class Arg1, class Arg2>
  static void do_task2(const arg_when_2<Empty, Func, Arg1, Arg2>& args,
		       const Callback::SafeFunctorArg<const Ret&>& when) {
    Callback::post(Callback::make_ref(&do_when<Ret>,
                                      when,
                                      args.func(args.arg1, args.arg2)),
                   args.priority,
                   args.context);
    // if this callback doesn't throw so post_fail() doesn't execute,
    // we need to delete do_fail_cb.  Destructors of bound arguments
    // of do_fail_cb are not supposed to throw, but we might as well
    // deal with it here.  An elipsis catch block is fine as thread
    // cancellation is blocked in all tasks.
    try {
      delete args.do_fail_cb;
    }
    catch (...) {}
  }

  template <class Ret>
  static void do_when(const Callback::SafeFunctorArg<const Ret&>& when,
                      const Ret& val) {
    when(val);
  }

  template <class Func, class Ret>
  static void do_task_rel0(const arg_when_0<Empty, Func>& args,
			   const SharedLockPtr<SafeEmitterArg<const Ret&> >& when_emitter) {
    Callback::post(Callback::make_ref(&do_when_rel<Ret>,
                                      when_emitter,
                                      args.func()),
                   args.priority,
                   args.context);
    // if this callback doesn't throw so post_fail() doesn't execute,
    // we need to delete do_fail_cb.  Destructors of bound arguments
    // of do_fail_cb are not supposed to throw, but we might as well
    // deal with it here.  An elipsis catch block is fine as thread
    // cancellation is blocked in all tasks.
    try {
      delete args.do_fail_cb;
    }
    catch (...) {}
  }

  template <class Func, class Ret, class Arg1>
  static void do_task_rel1(const arg_when_1<Empty, Func, Arg1>& args,
			   const SharedLockPtr<SafeEmitterArg<const Ret&> >& when_emitter) {
    Callback::post(Callback::make_ref(&do_when_rel<Ret>,
                                      when_emitter,
                                      args.func(args.arg1)),
                   args.priority,
                   args.context);
    // if this callback doesn't throw so post_fail() doesn't execute,
    // we need to delete do_fail_cb.  Destructors of bound arguments
    // of do_fail_cb are not supposed to throw, but we might as well
    // deal with it here.  An elipsis catch block is fine as thread
    // cancellation is blocked in all tasks.
    try {
      delete args.do_fail_cb;
    }
    catch (...) {}
  }

  template <class Func, class Ret, class Arg1, class Arg2>
  static void do_task_rel2(const arg_when_2<Empty, Func, Arg1, Arg2>& args,
			   const SharedLockPtr<SafeEmitterArg<const Ret&> >& when_emitter) {
    Callback::post(Callback::make_ref(&do_when_rel<Ret>,
                                      when_emitter,
                                      args.func(args.arg1, args.arg2)),
                   args.priority,
                   args.context);
    // if this callback doesn't throw so post_fail() doesn't execute,
    // we need to delete do_fail_cb.  Destructors of bound arguments
    // of do_fail_cb are not supposed to throw, but we might as well
    // deal with it here.  An elipsis catch block is fine as thread
    // cancellation is blocked in all tasks.
    try {
      delete args.do_fail_cb;
    }
    catch (...) {}
  }

  template <class Ret>
  static void do_when_rel(const SharedLockPtr<SafeEmitterArg<const Ret&> >& when_emitter,
                          const Ret& val) {
    when_emitter->emit(val);
  }

  static void post_fail(const Callback::Callback* do_fail_cb,
                        std::pair<gint, GMainContext*> loop_args) {
    // this function does not throw
    Callback::post(do_fail_cb, loop_args.first, loop_args.second);
  }

  static void do_fail_rel(const SharedLockPtr<SafeEmitter>& fail_emitter) {
    fail_emitter->emit(); // might throw std::bad_alloc
  }
};

// the 'Ret' type has to be specifically stated: it cannot be deduced
// from the arguments.  We pass Func as a template type so we do not
// have to duplicate code for const and non-const object methods (this
// is instead represented by the Func type)
template <class Ret, class T, class Func>
Cgu::SharedLockPtr<Cgu::AsyncResult<Ret> > memfun_result_0(TaskManager& tm,
							   T& t,
							   Func func) {
  SharedLockPtr<AsyncResult<Ret> > ret(new AsyncResult<Ret>);

  typedef std::auto_ptr<const Callback::Callback> cb_ptr;

  cb_ptr exec_cb(Callback::make_ref(&MemfunResultWrapper::exec0<T&, Func, Ret>,
				    arg_result_0<T&, Func>(t,
							   func),
				    ret));
  cb_ptr do_fail_cb(Callback::make_ref(&MemfunResultWrapper::do_fail<Ret>,
				       ret));
  tm.add_task(exec_cb, do_fail_cb);

  return ret;
}

// the 'Ret' and 'Arg1' types have to be specifically stated: they
// cannot be deduced from the arguments.  We pass Func as a template
// type so we do not have to duplicate code for const and non-const
// object methods (this is instead represented by the Func type)
template <class Ret, class Arg1, class T, class Func>
Cgu::SharedLockPtr<Cgu::AsyncResult<Ret> > memfun_result_1(TaskManager& tm,
							   T& t,
							   Func func,
							   typename Cgu::Param<Arg1>::ParamType arg1) {
  SharedLockPtr<AsyncResult<Ret> > ret(new AsyncResult<Ret>);

  typedef std::auto_ptr<const Callback::Callback> cb_ptr;

  cb_ptr exec_cb(Callback::make_ref(&MemfunResultWrapper::exec1<T&, Func, Ret, Arg1>,
				    arg_result_1<T&, Func, Arg1>(t,
								 func,
								 arg1),
				    ret));
  cb_ptr do_fail_cb(Callback::make_ref(&MemfunResultWrapper::do_fail<Ret>,
				       ret));
  tm.add_task(exec_cb, do_fail_cb);

  return ret;
}

// the 'Ret', 'Arg1' and 'Arg2' types have to be specifically stated:
// they cannot be deduced from the arguments.  We pass Func as a
// template type so we do not have to duplicate code for const and
// non-const object methods (this is instead represented by the Func
// type)
template <class Ret, class Arg1, class Arg2, class T, class Func>
Cgu::SharedLockPtr<Cgu::AsyncResult<Ret> > memfun_result_2(TaskManager& tm,
							   T& t,
							   Func func,
							   typename Cgu::Param<Arg1>::ParamType arg1,
							   typename Cgu::Param<Arg2>::ParamType arg2) {
  SharedLockPtr<AsyncResult<Ret> > ret(new AsyncResult<Ret>);

  typedef std::auto_ptr<const Callback::Callback> cb_ptr;

  cb_ptr exec_cb(Callback::make_ref(&MemfunResultWrapper::exec2<T&, Func, Ret, Arg1, Arg2>,
				    arg_result_2<T&, Func, Arg1, Arg2>(t,
								       func,
								       arg1,
								       arg2),
				    ret));
  cb_ptr do_fail_cb(Callback::make_ref(&MemfunResultWrapper::do_fail<Ret>,
				       ret));
  tm.add_task(exec_cb, do_fail_cb);

  return ret;
}

// the 'Ret' type has to be specifically stated: it cannot be deduced
// from the arguments.  We pass Func as a template type so we do not
// have to duplicate code for const and non-const object methods (this
// is instead represented by the Func type)
template <class Ret, class T, class Func>
void memfun_when_0(TaskManager& tm,
		   std::auto_ptr<const Cgu::Callback::CallbackArg<const Ret&> > when,
		   Cgu::Releaser* when_releaser,
		   std::auto_ptr<const Cgu::Callback::Callback> fail,
		   Cgu::Releaser* fail_releaser,
		   gint priority,
		   GMainContext* context,
		   T& t,
		   Func func) {
 
  // we construct the callback for MemfunWhenWrapper::do_fail() here
  // rather than in MemfunWhenWrapper::post_fail(), so that
  // MemfunWhenWrapper::post_fail() cannot throw
  const Callback::Callback* do_fail_cb = 0;
  if (fail.get()) {
    if (fail_releaser) {
      SharedLockPtr<SafeEmitter> fail_emitter(new SafeEmitter);
      fail_emitter->connect(fail.release(), *fail_releaser);
      do_fail_cb = Callback::make_ref(&MemfunWhenWrapper::do_fail_rel,
                                      fail_emitter);
    }
    else {
      do_fail_cb = fail.release();
    }
  }

  try {
    std::auto_ptr<const Callback::Callback> post_fail_cb;
    if (do_fail_cb) {
      post_fail_cb.reset(Callback::make_ref(&MemfunWhenWrapper::post_fail,
					    do_fail_cb,
					    std::make_pair(priority, context)));
    }

    std::auto_ptr<const Callback::Callback> do_task_cb;
    if (when_releaser) {
      SharedLockPtr<SafeEmitterArg<const Ret&> > when_emitter(new SafeEmitterArg<const Ret&>);
      when_emitter->connect(when.release(), *when_releaser);
      do_task_cb.reset(Callback::make_ref(&MemfunWhenWrapper::do_task_rel0<T&, Func, Ret>,
					  arg_when_0<T&, Func>(t,
							       func,
							       do_fail_cb,
							       priority,
							       context),
					  when_emitter));
    }
    else {
      Callback::SafeFunctorArg<const Ret&> when_functor(when.release());
      do_task_cb.reset(Callback::make_ref(&MemfunWhenWrapper::do_task0<T&, Func, Ret>,
					  arg_when_0<T&, Func>(t,
							       func,
							       do_fail_cb,
							       priority,
							       context),
					  
					  when_functor));
    }
    tm.add_task(do_task_cb, post_fail_cb);
  }
  catch (...) {
    // Destructors of bound arguments of do_fail_cb are not supposed
    // to throw, but we might as well deal with it here.
    CancelBlock b;
    try {
      delete do_fail_cb;
    }
    catch (...) {}
    throw;
  }
}

// the 'Ret' and 'Arg1' types have to be specifically stated: they
// cannot be deduced from the arguments.  We pass Func as a template
// type so we do not have to duplicate code for const and non-const
// object methods (this is instead represented by the Func type)
template <class Ret, class Arg1, class T, class Func>
void memfun_when_1(TaskManager& tm,
		   std::auto_ptr<const Cgu::Callback::CallbackArg<const Ret&> > when,
		   Cgu::Releaser* when_releaser,
		   std::auto_ptr<const Cgu::Callback::Callback> fail,
		   Cgu::Releaser* fail_releaser,
		   gint priority,
		   GMainContext* context,
		   T& t,
		   Func func,
		   typename Cgu::Param<Arg1>::ParamType arg1) {

  // we construct the callback for MemfunWhenWrapper::do_fail() here
  // rather than in MemfunWhenWrapper::post_fail(), so that
  // MemfunWhenWrapper::post_fail() cannot throw
  const Callback::Callback* do_fail_cb = 0;
  if (fail.get()) {
    if (fail_releaser) {
      SharedLockPtr<SafeEmitter> fail_emitter(new SafeEmitter);
      fail_emitter->connect(fail.release(), *fail_releaser);
      do_fail_cb = Callback::make_ref(&MemfunWhenWrapper::do_fail_rel,
                                      fail_emitter);
    }
    else {
      do_fail_cb = fail.release();
    }
  }

  try {
    std::auto_ptr<const Callback::Callback> post_fail_cb;
    if (do_fail_cb) {
      post_fail_cb.reset(Callback::make_ref(&MemfunWhenWrapper::post_fail,
					    do_fail_cb,
					    std::make_pair(priority, context)));
    }

    std::auto_ptr<const Callback::Callback> do_task_cb;
    if (when_releaser) {
      SharedLockPtr<SafeEmitterArg<const Ret&> > when_emitter(new SafeEmitterArg<const Ret&>);
      when_emitter->connect(when.release(), *when_releaser);
      do_task_cb.reset(Callback::make_ref(&MemfunWhenWrapper::do_task_rel1<T&, Func, Ret, Arg1>,
					  arg_when_1<T&, Func, Arg1>(t,
								     func,
								     do_fail_cb,
								     priority,
								     context,
								     arg1),
					  when_emitter));
    }
    else {
      Callback::SafeFunctorArg<const Ret&> when_functor(when.release());
      do_task_cb.reset(Callback::make_ref(&MemfunWhenWrapper::do_task1<T&, Func, Ret, Arg1>,
					  arg_when_1<T&, Func, Arg1>(t,
								     func,
								     do_fail_cb,
								     priority,
								     context,
								     arg1),
					  when_functor));
    }
    tm.add_task(do_task_cb, post_fail_cb);
  }
  catch (...) {
    // Destructors of bound arguments of do_fail_cb are not supposed
    // to throw, but we might as well deal with it here.
    CancelBlock b;
    try {
      delete do_fail_cb;
    }
    catch (...) {}
    throw;
  }
}

// the 'Ret', 'Arg1' and 'Arg2' types have to be specifically stated:
// they cannot be deduced from the arguments.  We pass Func as a
// template type so we do not have to duplicate code for const and
// non-const object methods (this is instead represented by the Func
// type)
template <class Ret, class Arg1, class Arg2, class T, class Func>
void memfun_when_2(TaskManager& tm,
		   std::auto_ptr<const Cgu::Callback::CallbackArg<const Ret&> > when,
		   Cgu::Releaser* when_releaser,
		   std::auto_ptr<const Cgu::Callback::Callback> fail,
		   Cgu::Releaser* fail_releaser,
		   gint priority,
		   GMainContext* context,
		   T& t,
		   Func func,
		   typename Cgu::Param<Arg1>::ParamType arg1,
		   typename Cgu::Param<Arg2>::ParamType arg2) {

  // we construct the callback for MemfunWhenWrapper::do_fail() here
  // rather than in MemfunWhenWrapper::post_fail(), so that
  // MemfunWhenWrapper::post_fail() cannot throw
  const Callback::Callback* do_fail_cb = 0;
  if (fail.get()) {
    if (fail_releaser) {
      SharedLockPtr<SafeEmitter> fail_emitter(new SafeEmitter);
      fail_emitter->connect(fail.release(), *fail_releaser);
      do_fail_cb = Callback::make_ref(&MemfunWhenWrapper::do_fail_rel,
                                      fail_emitter);
    }
    else {
      do_fail_cb = fail.release();
    }
  }

  try {
    std::auto_ptr<const Callback::Callback> post_fail_cb;
    if (do_fail_cb) {
      post_fail_cb.reset(Callback::make_ref(&MemfunWhenWrapper::post_fail,
					    do_fail_cb,
					    std::make_pair(priority, context)));
    }

    std::auto_ptr<const Callback::Callback> do_task_cb;
    if (when_releaser) {
      SharedLockPtr<SafeEmitterArg<const Ret&> > when_emitter(new SafeEmitterArg<const Ret&>);
      when_emitter->connect(when.release(), *when_releaser);
      do_task_cb.reset(Callback::make_ref(&MemfunWhenWrapper::do_task_rel2<T&, Func, Ret, Arg1, Arg2>,
					  arg_when_2<T&, Func,
					             Arg1, Arg2>(t,
								 func,
								 do_fail_cb,
								 priority,
								 context,
								 arg1,
								 arg2),
					  when_emitter));
    }
    else {
      Callback::SafeFunctorArg<const Ret&> when_functor(when.release());
      do_task_cb.reset(Callback::make_ref(&MemfunWhenWrapper::do_task2<T&, Func, Ret, Arg1, Arg2>,
					  arg_when_2<T&, Func,
					             Arg1, Arg2>(t,
								 func,
								 do_fail_cb,
								 priority,
								 context,
								 arg1,
								 arg2),
					  when_functor));
    }
    tm.add_task(do_task_cb, post_fail_cb);
  }
  catch (...) {
    // Destructors of bound arguments of do_fail_cb are not supposed
    // to throw, but we might as well deal with it here.
    CancelBlock b;
    try {
      delete do_fail_cb;
    }
    catch (...) {}
    throw;
  }
}

} // namespace TaskManagerHelper

template <class Ret, class T>
Cgu::SharedLockPtr<Cgu::AsyncResult<Ret> > TaskManager::make_task_result(T& t,
									 Ret (T::*func)()) {
  return TaskManagerHelper::memfun_result_0<Ret>(*this, t, func);
}

template <class Ret, class Arg1, class T>
Cgu::SharedLockPtr<Cgu::AsyncResult<Ret> > TaskManager::make_task_result(T& t,
									 Ret (T::*func)(Arg1),
									 typename Cgu::Param<Arg1>::ParamType arg1) {
  return TaskManagerHelper::memfun_result_1<Ret, Arg1>(*this, t, func, arg1);
}

template <class Ret, class Arg1, class Arg2, class T>
Cgu::SharedLockPtr<Cgu::AsyncResult<Ret> > TaskManager::make_task_result(T& t,
									 Ret (T::*func)(Arg1, Arg2),
									 typename Cgu::Param<Arg1>::ParamType arg1,
									 typename Cgu::Param<Arg2>::ParamType arg2) {
  return TaskManagerHelper::memfun_result_2<Ret, Arg1, Arg2>(*this, t, func, arg1, arg2);
}

template <class Ret, class T>
Cgu::SharedLockPtr<Cgu::AsyncResult<Ret> > TaskManager::make_task_result(const T& t,
									 Ret (T::*func)() const) {
  return TaskManagerHelper::memfun_result_0<Ret>(*this, t, func);
}

template <class Ret, class Arg1, class T>
Cgu::SharedLockPtr<Cgu::AsyncResult<Ret> > TaskManager::make_task_result(const T& t,
									 Ret (T::*func)(Arg1) const,
									 typename Cgu::Param<Arg1>::ParamType arg1) {
  return TaskManagerHelper::memfun_result_1<Ret, Arg1>(*this, t, func, arg1);
}

template <class Ret, class Arg1, class Arg2, class T>
Cgu::SharedLockPtr<Cgu::AsyncResult<Ret> > TaskManager::make_task_result(const T& t,
									 Ret (T::*func)(Arg1, Arg2) const,
									 typename Cgu::Param<Arg1>::ParamType arg1,
									 typename Cgu::Param<Arg2>::ParamType arg2) {
  return TaskManagerHelper::memfun_result_2<Ret, Arg1, Arg2>(*this, t, func, arg1, arg2);
}

template <class Ret>
Cgu::SharedLockPtr<Cgu::AsyncResult<Ret> > TaskManager::make_task_result(Ret (*func)()) {

  SharedLockPtr<AsyncResult<Ret> > ret(new AsyncResult<Ret>);

  typedef Ret(*Func)();
  typedef std::auto_ptr<const Callback::Callback> cb_ptr;

  cb_ptr exec_cb(Callback::make_ref(&TaskManagerHelper::FunResultWrapper::exec0<Func, Ret>,
				    TaskManagerHelper::arg_result_0<TaskManagerHelper::Empty,
				                                    Func>(TaskManagerHelper::Empty(),
									  func),
				    ret));
  cb_ptr do_fail_cb(Callback::make_ref(&TaskManagerHelper::FunResultWrapper::do_fail<Ret>,
				       ret));
  add_task(exec_cb, do_fail_cb);

  return ret;
}

template <class Ret, class Arg1>
Cgu::SharedLockPtr<Cgu::AsyncResult<Ret> > TaskManager::make_task_result(Ret (*func)(Arg1),
									 typename Cgu::Param<Arg1>::ParamType arg1) {

  SharedLockPtr<AsyncResult<Ret> > ret(new AsyncResult<Ret>);

  typedef Ret(*Func)(Arg1);
  typedef std::auto_ptr<const Callback::Callback> cb_ptr;

  cb_ptr exec_cb(Callback::make_ref(&TaskManagerHelper::FunResultWrapper::exec1<Func, Ret, Arg1>,
				    TaskManagerHelper::arg_result_1<TaskManagerHelper::Empty,
				                                    Func,
				                                    Arg1>(TaskManagerHelper::Empty(),
									  func,
									  arg1),
				    ret));
  cb_ptr do_fail_cb(Callback::make_ref(&TaskManagerHelper::FunResultWrapper::do_fail<Ret>,
				       ret));
  add_task(exec_cb, do_fail_cb);

  return ret;
}

template <class Ret, class Arg1, class Arg2>
Cgu::SharedLockPtr<Cgu::AsyncResult<Ret> > TaskManager::make_task_result(Ret (*func)(Arg1, Arg2),
									 typename Cgu::Param<Arg1>::ParamType arg1,
									 typename Cgu::Param<Arg2>::ParamType arg2) {

  SharedLockPtr<AsyncResult<Ret> > ret(new AsyncResult<Ret>);

  typedef Ret(*Func)(Arg1, Arg2);
  typedef std::auto_ptr<const Callback::Callback> cb_ptr;

  cb_ptr exec_cb(Callback::make_ref(&TaskManagerHelper::FunResultWrapper::exec2<Func, Ret, Arg1, Arg2>,
				    TaskManagerHelper::arg_result_2<TaskManagerHelper::Empty,
				                                    Func,
				                                    Arg1,
				                                    Arg2>(TaskManagerHelper::Empty(),
									  func,
									  arg1,
									  arg2),
				    ret));
  cb_ptr do_fail_cb(Callback::make_ref(&TaskManagerHelper::FunResultWrapper::do_fail<Ret>,
				       ret));
  add_task(exec_cb, do_fail_cb);

  return ret;
}

template <class Ret, class T>
void TaskManager::make_task_when_full(std::auto_ptr<const Cgu::Callback::CallbackArg<const Ret&> > when,
				      Cgu::Releaser* when_releaser,
				      std::auto_ptr<const Cgu::Callback::Callback> fail,
				      Cgu::Releaser* fail_releaser,
				      gint priority,
				      GMainContext* context,
				      T& t,
				      Ret (T::*func)()) {
  
  TaskManagerHelper::memfun_when_0<Ret>(*this,
					when, when_releaser,
					fail, fail_releaser,
					priority, context,
					t, func);
}

template <class Ret, class Arg1, class T>
void TaskManager::make_task_when_full(std::auto_ptr<const Cgu::Callback::CallbackArg<const Ret&> > when,
				      Cgu::Releaser* when_releaser,
				      std::auto_ptr<const Cgu::Callback::Callback> fail,
				      Cgu::Releaser* fail_releaser,
				      gint priority,
				      GMainContext* context,
				      T& t,
				      Ret (T::*func)(Arg1),
				      typename Cgu::Param<Arg1>::ParamType arg1) {
  
  TaskManagerHelper::memfun_when_1<Ret, Arg1>(*this,
					      when, when_releaser,
					      fail, fail_releaser,
					      priority, context,
					      t, func, arg1);
}

template <class Ret, class Arg1, class Arg2, class T>
void TaskManager::make_task_when_full(std::auto_ptr<const Cgu::Callback::CallbackArg<const Ret&> > when,
				      Cgu::Releaser* when_releaser,
				      std::auto_ptr<const Cgu::Callback::Callback> fail,
				      Cgu::Releaser* fail_releaser,
				      gint priority,
				      GMainContext* context,
				      T& t,
				      Ret (T::*func)(Arg1, Arg2),
				      typename Cgu::Param<Arg1>::ParamType arg1,
				      typename Cgu::Param<Arg2>::ParamType arg2) {

  TaskManagerHelper::memfun_when_2<Ret, Arg1, Arg2>(*this,
						    when, when_releaser,
						    fail, fail_releaser,
						    priority, context,
						    t, func, arg1, arg2);
}

template <class Ret, class T>
void TaskManager::make_task_when_full(std::auto_ptr<const Cgu::Callback::CallbackArg<const Ret&> > when,
				      Cgu::Releaser* when_releaser,
				      std::auto_ptr<const Cgu::Callback::Callback> fail,
				      Cgu::Releaser* fail_releaser,
				      gint priority,
				      GMainContext* context,
				      const T& t,
				      Ret (T::*func)() const) {

  TaskManagerHelper::memfun_when_0<Ret>(*this,
					when, when_releaser,
					fail, fail_releaser,
					priority, context,
					t, func);
}

template <class Ret, class Arg1, class T>
void TaskManager::make_task_when_full(std::auto_ptr<const Cgu::Callback::CallbackArg<const Ret&> > when,
				      Cgu::Releaser* when_releaser,
				      std::auto_ptr<const Cgu::Callback::Callback> fail,
				      Cgu::Releaser* fail_releaser,
				      gint priority,
				      GMainContext* context,
				      const T& t,
				      Ret (T::*func)(Arg1) const,
				      typename Cgu::Param<Arg1>::ParamType arg1) {
  
  TaskManagerHelper::memfun_when_1<Ret, Arg1>(*this,
					      when, when_releaser,
					      fail, fail_releaser,
					      priority, context,
					      t, func, arg1);
}

template <class Ret, class Arg1, class Arg2, class T>
void TaskManager::make_task_when_full(std::auto_ptr<const Cgu::Callback::CallbackArg<const Ret&> > when,
				      Cgu::Releaser* when_releaser,
				      std::auto_ptr<const Cgu::Callback::Callback> fail,
				      Cgu::Releaser* fail_releaser,
				      gint priority,
				      GMainContext* context,
				      const T& t,
				      Ret (T::*func)(Arg1, Arg2) const,
				      typename Cgu::Param<Arg1>::ParamType arg1,
				      typename Cgu::Param<Arg2>::ParamType arg2) {

  TaskManagerHelper::memfun_when_2<Ret, Arg1, Arg2>(*this,
						    when, when_releaser,
						    fail, fail_releaser,
						    priority, context,
						    t, func, arg1, arg2);
}

template <class Ret>
void TaskManager::make_task_when_full(std::auto_ptr<const Cgu::Callback::CallbackArg<const Ret&> > when,
				      Cgu::Releaser* when_releaser,
				      std::auto_ptr<const Cgu::Callback::Callback> fail,
				      Cgu::Releaser* fail_releaser,
				      gint priority,
				      GMainContext* context,
				      Ret (*func)()) {

  typedef Ret(*Func)();

  // we construct the callback for FunWhenWrapper::do_fail() here
  // rather than in FunWhenWrapper::post_fail(), so that
  // FunWhenWrapper::post_fail() cannot throw
  const Callback::Callback* do_fail_cb = 0;
  if (fail.get()) {
    if (fail_releaser) {
      SharedLockPtr<SafeEmitter> fail_emitter(new SafeEmitter);
      fail_emitter->connect(fail.release(), *fail_releaser);
      do_fail_cb = Callback::make_ref(&TaskManagerHelper::FunWhenWrapper::do_fail_rel,
                                      fail_emitter);
    }
    else {
      do_fail_cb = fail.release();
    }
  }

  try {
    std::auto_ptr<const Callback::Callback> post_fail_cb;
    if (do_fail_cb) {
      post_fail_cb.reset(Callback::make_ref(&TaskManagerHelper::FunWhenWrapper::post_fail,
					    do_fail_cb,
					    std::make_pair(priority, context)));
    }

    std::auto_ptr<const Callback::Callback> do_task_cb;
    if (when_releaser) {
        SharedLockPtr<SafeEmitterArg<const Ret&> > when_emitter(new SafeEmitterArg<const Ret&>);
        when_emitter->connect(when.release(), *when_releaser);;
	do_task_cb.reset(Callback::make_ref(&TaskManagerHelper::FunWhenWrapper::do_task_rel0<Func, Ret>,
					    TaskManagerHelper::arg_when_0<TaskManagerHelper::Empty,
					                                  Func>(TaskManagerHelper::Empty(),
										func,
										do_fail_cb,
										priority,
										context),
					    when_emitter));
    }
    else {
      Callback::SafeFunctorArg<const Ret&> when_functor(when.release());
      do_task_cb.reset(Callback::make_ref(&TaskManagerHelper::FunWhenWrapper::do_task0<Func, Ret>,
					  TaskManagerHelper::arg_when_0<TaskManagerHelper::Empty,
					                                Func>(TaskManagerHelper::Empty(),
									      func,
									      do_fail_cb,
									      priority,
									      context),
					  when_functor));
    }
    add_task(do_task_cb, post_fail_cb);
  }
  catch (...) {
    // Destructors of bound arguments of do_fail_cb are not supposed
    // to throw, but we might as well deal with it here.
    CancelBlock b;
    try {
      delete do_fail_cb;
    }
    catch (...) {}
    throw;
  }
}

template <class Ret, class Arg1>
void TaskManager::make_task_when_full(std::auto_ptr<const Cgu::Callback::CallbackArg<const Ret&> > when,
				      Cgu::Releaser* when_releaser,
				      std::auto_ptr<const Cgu::Callback::Callback> fail,
				      Cgu::Releaser* fail_releaser,
				      gint priority,
				      GMainContext* context,
				      Ret (*func)(Arg1),
				      typename Cgu::Param<Arg1>::ParamType arg1) {

  typedef Ret(*Func)(Arg1);

  // we construct the callback for FunWhenWrapper::do_fail() here
  // rather than in FunWhenWrapper::post_fail(), so that
  // FunWhenWrapper::post_fail() cannot throw
  const Callback::Callback* do_fail_cb = 0;
  if (fail.get()) {
    if (fail_releaser) {
      SharedLockPtr<SafeEmitter> fail_emitter(new SafeEmitter);
      fail_emitter->connect(fail.release(), *fail_releaser);
      do_fail_cb = Callback::make_ref(&TaskManagerHelper::FunWhenWrapper::do_fail_rel,
                                      fail_emitter);
    }
    else {
      do_fail_cb = fail.release();
    }
  }

  try {
    std::auto_ptr<const Callback::Callback> post_fail_cb;
    if (do_fail_cb) {
      post_fail_cb.reset(Callback::make_ref(&TaskManagerHelper::FunWhenWrapper::post_fail,
					    do_fail_cb,
					    std::make_pair(priority, context)));
    }

    std::auto_ptr<const Callback::Callback> do_task_cb;
    if (when_releaser) {
      SharedLockPtr<SafeEmitterArg<const Ret&> > when_emitter(new SafeEmitterArg<const Ret&>);
      when_emitter->connect(when.release(), *when_releaser);
      do_task_cb.reset(Callback::make_ref(&TaskManagerHelper::FunWhenWrapper::do_task_rel1<Func, Ret, Arg1>,
					  TaskManagerHelper::arg_when_1<TaskManagerHelper::Empty,
					                                Func,
					                                Arg1>(TaskManagerHelper::Empty(),
									      func,
									      do_fail_cb,
									      priority,
									      context,
									      arg1),
					  when_emitter));
    }
    else {
	Callback::SafeFunctorArg<const Ret&> when_functor(when.release());
	do_task_cb.reset(Callback::make_ref(&TaskManagerHelper::FunWhenWrapper::do_task1<Func, Ret, Arg1>,
					    TaskManagerHelper::arg_when_1<TaskManagerHelper::Empty,
					                                  Func,
					                                  Arg1>(TaskManagerHelper::Empty(),
										func,
										do_fail_cb,
										priority,
										context,
										arg1),
					    when_functor));
    }
    add_task(do_task_cb, post_fail_cb);
  }
  catch (...) {
    // Destructors of bound arguments of do_fail_cb are not supposed
    // to throw, but we might as well deal with it here.
    CancelBlock b;
    try {
      delete do_fail_cb;
    }
    catch (...) {}
    throw;
  }
}

template <class Ret, class Arg1, class Arg2>
void TaskManager::make_task_when_full(std::auto_ptr<const Cgu::Callback::CallbackArg<const Ret&> > when,
				      Cgu::Releaser* when_releaser,
				      std::auto_ptr<const Cgu::Callback::Callback> fail,
				      Cgu::Releaser* fail_releaser,
				      gint priority,
				      GMainContext* context,
				      Ret (*func)(Arg1, Arg2),
				      typename Cgu::Param<Arg1>::ParamType arg1,
				      typename Cgu::Param<Arg2>::ParamType arg2) {

  typedef Ret(*Func)(Arg1, Arg2);

  // we construct the callback for FunWhenWrapper::do_fail() here
  // rather than in FunWhenWrapper::post_fail(), so that
  // FunWhenWrapper::post_fail() cannot throw
  const Callback::Callback* do_fail_cb = 0;
  if (fail.get()) {
    if (fail_releaser) {
      SharedLockPtr<SafeEmitter> fail_emitter(new SafeEmitter);
      fail_emitter->connect(fail.release(), *fail_releaser);
      do_fail_cb = Callback::make_ref(&TaskManagerHelper::FunWhenWrapper::do_fail_rel,
                                      fail_emitter);
    }
    else {
      do_fail_cb = fail.release();
    }
  }

  try {
    std::auto_ptr<const Callback::Callback> post_fail_cb;
    if (do_fail_cb) {
      post_fail_cb.reset(Callback::make_ref(&TaskManagerHelper::FunWhenWrapper::post_fail,
					    do_fail_cb,
					    std::make_pair(priority, context)));
    }

    std::auto_ptr<const Callback::Callback> do_task_cb;
    if (when_releaser) {
      SharedLockPtr<SafeEmitterArg<const Ret&> > when_emitter(new SafeEmitterArg<const Ret&>);
      when_emitter->connect(when.release(), *when_releaser);
      do_task_cb.reset(Callback::make_ref(&TaskManagerHelper::FunWhenWrapper::do_task_rel2<Func, Ret, Arg1, Arg2>,
					  TaskManagerHelper::arg_when_2<TaskManagerHelper::Empty,
					                                Func,
					                                Arg1,
					                                Arg2>(TaskManagerHelper::Empty(),
									      func,
									      do_fail_cb,
									      priority,
									      context,
									      arg1,
									      arg2),
					  when_emitter));
    }
    else {
      Callback::SafeFunctorArg<const Ret&> when_functor(when.release());
      do_task_cb.reset(Callback::make_ref(&TaskManagerHelper::FunWhenWrapper::do_task2<Func, Ret, Arg1, Arg2>,
					  TaskManagerHelper::arg_when_2<TaskManagerHelper::Empty,
					                                Func,
					                                Arg1,
					                                Arg2>(TaskManagerHelper::Empty(),
									      func,
									      do_fail_cb,
									      priority,
									      context,
									      arg1,
									      arg2),
					  when_functor));
    }
    add_task(do_task_cb, post_fail_cb);
  }
  catch (...) {
    // Destructors of bound arguments of do_fail_cb are not supposed
    // to throw, but we might as well deal with it here.
    CancelBlock b;
    try {
      delete do_fail_cb;
    }
    catch (...) {}
    throw;
  }
}

} // namespace Thread

} // namespace Cgu

#endif // CGU_TASK_MANAGER_TPP
